"""Searching python object docs

You know when you're using a new package, you REPL a few things, and you wonder what
you can do with an object you're holding in your hand.

You can go look online, but it's not always that easy to find the "what can I do with
this particular object" answers.

You can do `help(obj)` and browse it's output.

... or, you can do what we're showing here.


docs_searcher: Search the docs of the attributes of the input python object.

    >>> from grub import TfidfKnnSearcher
    >>> search = docs_searcher(TfidfKnnSearcher, n_results=2)
    >>> len(search)
    6
    >>> keys = search('make index')
    >>> list(keys)
    ['fit', 'fit_knn']

    Those are the names of the top 2 (=n_results) methods matching the query 'make
    index'. You can see the actual docs using `[...]`:

    >>> print(search['fit'])
    Fit all models of pipeline, i.e. making a search index.

Note: There's a similar function in `grub.pycode` called `search_documented_attributes`

"""

from sklearn.neighbors import NearestNeighbors

from grub import TfidfKnnSearcher


def not_dundered(string):
    return not string.startswith('__')


def not_underscored(string):
    return not string.startswith('_')


DFLT_ATTRNAME_FILT = not_underscored
DFLT_ATTR_FILT = callable


def attr_docs(obj, attrname_filt=DFLT_ATTRNAME_FILT, attr_filt=DFLT_ATTR_FILT):
    for attrname in filter(attrname_filt, dir(obj)):
        attr = getattr(obj, attrname)
        if attr_filt(attr):
            doc = getattr(attr, '__doc__', None)
            if doc:
                yield attrname, doc


def docs_searcher(
    obj, n_results=10, attrname_filt=DFLT_ATTRNAME_FILT, attr_filt=DFLT_ATTR_FILT
):
    """Search the docs of the attributes of the input python object.

    >>> from grub import TfidfKnnSearcher
    >>> search = docs_searcher(TfidfKnnSearcher, n_results=2)
    >>> len(search)
    6
    >>> keys = search('make index')
    >>> list(keys)
    ['fit', 'fit_knn']

    Those are the names of the top 2 (=n_results) methods matching the query 'make
    index'. You can see the actual docs using `[...]`:

    >>> print(search['fit'])
    Fit all models of pipeline, i.e. making a search index.

    """
    docs = dict(attr_docs(obj, attrname_filt, attr_filt))
    knn = NearestNeighbors(n_neighbors=n_results, metric='cosine')
    return TfidfKnnSearcher(docs, knn=knn).fit()



"""Base functionality"""

import re
from typing import Mapping, Union
from types import ModuleType
from dataclasses import dataclass
from inspect import ismodule
import os

import numpy as np
from sklearn.feature_extraction.text import TfidfVectorizer
from sklearn.neighbors import NearestNeighbors
from sklearn.utils.validation import check_is_fitted

from py2store import LocalTextStore, cached_keys, lazyprop, KvReader


def grub(search_store, query, n=10):
    search_store = get_py_files_store(search_store)
    knn = NearestNeighbors(n_neighbors=n, metric='cosine')
    search = TextFilesSearcher(search_store, knn=knn).fit()
    return search(query)


def get_py_files_store(spec):
    if ismodule(spec):
        spec = os.path.dirname(spec.__file__)
    if isinstance(spec, str):
        if spec.endswith('.py'):
            if spec.endswith('__init__.py'):
                spec = spec[: -len('__init__.py')]
            else:
                raise ValueError(
                    f'You provided an individual module. '
                    f'Must be a package module: {spec}'
                )
        if not spec.endswith(os.path.sep):
            spec = spec + os.path.sep
        spec = LocalTextStore(spec + '{}.py')
    return spec


camelcase_p = re.compile(
    r'''
    # Find words in a string. Order matters!
    [A-Z]+(?=[A-Z][a-z]) |  # All upper case before a capitalized word
    [A-Z]?[a-z]+ |  # Capitalized words / all lower case
    [A-Z]+ |  # All upper case
    \d+  # Numbers  TODO: Might want to drop that.
''',
    re.VERBOSE,
)


def camelcase_and_underscore_tokenizer(string):
    """
    >>> print(*camelcase_and_underscore_tokenizer('thereIsSomething i_wanted toShow'))
    there is something i wanted to show
    """
    return list(map(str.lower, camelcase_p.findall(string)))


class SearchStoreMixin(KvReader):
    store_attr = 'search_store'

    def __getitem__(self, k):
        return getattr(self, self.store_attr).__getitem__(k)

    def __iter__(self):
        return getattr(self, self.store_attr).__iter__()

    def __contains__(self, k):
        return getattr(self, self.store_attr).__contains__(k)

    def __len__(self):
        return getattr(self, self.store_attr).__len__()


class StoreMixin(SearchStoreMixin):
    store_attr = 'store'


class SearchStore(StoreMixin):
    """Build a search index for anything (that is given a mapping interface with string values).

    A store is anything with a ``collections.Mapping`` interface.
    Typically, a store's backend comes from local files or data-base wrapped into a mapping
    (see ``py2store`` for tools to do that!).
    For testing purposes though, we'll use a ``dict`` here:

    >>> store = {
    ... "Nelson Mandela": "The greatest glory in living lies not in never falling, but in rising every time we fall.",
    ... "Walt Disney": "The way to get started is to quit talking and begin doing.",
    ... "Steve Jobs": "Your time is limited, so don't waste it living someone else's life."
    ... }
    >>> search = SearchStore(store, n_neighbors=2)  # our store is small, so need to restrict our result size to less
    >>> list(search('living'))
    ['Steve Jobs', 'Nelson Mandela']

    A ``SearchStore`` instance is picklable.

    >>> import pickle
    >>> unserialized_search = pickle.loads(pickle.dumps(search))
    >>> list(unserialized_search('living'))
    ['Steve Jobs', 'Nelson Mandela']

    """

    _state_attrs = ('keys_array', 'tfidf', 'knn')

    def __init__(
        self, store, n_neighbors: int = 10, tokenizer=camelcase_and_underscore_tokenizer
    ):
        """Index store (values) and provide a search functionality for it.

        :param store: The collection that should be indexed for search.
            A store is anything with a ``collections.Mapping`` interface.
            Typically, a store's backend comes from local files or data-base wrapped into a mapping
            (see ``py2store`` for tools to do that!).
            If given as a string, it will be considered to be a file path and given to
            ``py2store.LocalTextStore`` to make a store.
        :param n_neighbors: The number search results to provide
        :param tokenizer: The function to apply to store (string) values to get an iterable of tokens.
            It is these tokens that are used both as the vocabulary for the queries and to represent searched values.
        """
        if isinstance(store, str):
            store = LocalTextStore(store)
        self.store = store

        self.tfidf = TfidfVectorizer(tokenizer=tokenizer, token_pattern=None)
        self.tfidf.fit(raw_documents=self.store.values())

        self.keys_array = np.array(list(self.store), dtype=object)

        doc_vecs = self.tfidf.fit_transform(self.store.values())
        self.n_neighbors = n_neighbors
        self.knn = NearestNeighbors(n_neighbors=n_neighbors, metric='cosine').fit(
            doc_vecs
        )

    def __getstate__(self):
        return {attr: getattr(self, attr) for attr in self._state_attrs}

    def __setstate__(self, d):
        for attr in self._state_attrs:
            setattr(self, attr, d[attr])

    def __call__(self, query):
        (score, *_), (idx, *_) = self.knn.kneighbors(self.tfidf.transform([query]))
        return self.keys_array[idx]

    def __getitem__(self, k):
        return self.store.__getitem__(k)


class DfltSearchStore(Mapping):
    def __iter__(self):
        yield from ()

    def __len__(self):
        return 0

    def __contains__(self):
        return False

    def __getitem__(self, item):
        raise KeyError('DfltSearchStore is not meant to be used')


class TfidfKnnFitMixin:
    def set_search_store(
        self, search_store
    ):  # TODO: override setter with this function
        self.search_store = search_store
        assert isinstance(self.search_store, Mapping), "Your store isn't a Mapping"
        return self.fit_knn()

    def fit_knn(self):
        """Fit the knn model"""
        self.knn.fit(self.fvs())
        return self

    def fit_tfidf(self, learn_store=None, error_callback=None):
        """Fit the tfidf model"""
        if learn_store is None:
            learn_store = self.search_store
        keys_fit_on_ = []
        raw_documents = iterate_values_and_accumulate_non_error_keys(
            learn_store, cache_keys_here=keys_fit_on_, error_callback=error_callback
        )
        self.tfidf.fit(raw_documents=raw_documents)
        self.tfidf.keys_fit_on_ = keys_fit_on_
        return self

    def fit(self, learn_store=None):
        """Fit all models of pipeline, i.e. making a search index."""
        return self.fit_tfidf(learn_store).fit_knn()


# search_store and
@dataclass
class TfidfKnnSearcher(SearchStoreMixin, TfidfKnnFitMixin):
    search_store: Mapping = DfltSearchStore()
    tfidf: TfidfVectorizer = TfidfVectorizer()
    knn: NearestNeighbors = NearestNeighbors(n_neighbors=10, metric='cosine')

    def __getattr__(self, attr):
        """Delegate method to wrapped store if not part of wrapper store methods"""
        return getattr(self.search_store, attr)

    @lazyprop
    def keys_array(self):
        return np.array(self.search_store)

    @property
    def n_neighbors(self):
        return self.knn.n_neighbors

    def fvs(self, error_callback=None):
        """The feature vectors for the search_store"""
        if hasattr(self.search_store, 'keys_cache'):
            return self.tfidf
        else:
            keys_cache = []
            search_documents = iterate_values_and_accumulate_non_error_keys(
                self.search_store,
                cache_keys_here=keys_cache,
                error_callback=error_callback,
            )
            fvs = self.tfidf.fit_transform(search_documents)
            self.search_store = cached_keys(self.search_store, keys_cache=keys_cache)
            return fvs

    # def fit(self, learn_store, ):
    def __call__(self, query):
        (_, *_), (idx, *_) = self.knn.kneighbors(self.tfidf.transform([query]))
        # # equivalently:
        # fv = self.query_to_fv(query)
        # idx, scores = self.fv_to_idx_and_score(fv)
        return self.keys_array[idx]

    def docs_to_fvs(self, docs):
        self.tfidf.transform(docs)

    def query_to_fv(self, query):
        return self.tfidf.transform([query])[0]

    def fv_to_idx_and_scores(self, fv):
        (scores, *_), (idx, *_) = self.knn.kneighbors([fv])
        return idx, scores

    _state_attrs = ('keys_array', 'tfidf', 'knn')

    def __getstate__(self):
        return {attr: getattr(self, attr) for attr in self._state_attrs}

    def __setstate__(self, d):
        for attr in self._state_attrs:
            setattr(self, attr, d[attr])


@dataclass
class TextFilesSearcherBase(TfidfKnnSearcher):
    search_store: Union[str, Mapping] = DfltSearchStore()
    tfidf: TfidfVectorizer = TfidfVectorizer(token_pattern=camelcase_p)

    def __post_init__(self):
        if isinstance(self.search_store, str):
            self.search_store = LocalTextStore(self.search_store)


@dataclass
class CodeSearcherBase(TextFilesSearcherBase):
    search_store: Union[str, Mapping, ModuleType] = DfltSearchStore()

    def __post_init__(self):
        self.search_store = get_py_files_store(self.search_store)


def iterate_values_and_accumulate_non_error_keys(
    store, cache_keys_here: list, errors_caught=Exception, error_callback=None
):
    for k in store:
        try:
            v = store[k]
            cache_keys_here.append(k)
            yield v
        except errors_caught as err:
            if error_callback is not None:
                error_callback(store, k, err)


# TODO: These do no more than their bases. Refactor.
class TextFilesSearcher(TextFilesSearcherBase):
    """Fittable Text searcher with fitters"""


class CodeSearcher(CodeSearcherBase):
    """Text searcher with fitters"""



"""Make a search object in the simplest way possible"""

from grub.base import (
    SearchStore,
    CodeSearcher,
    TfidfKnnSearcher,
    grub,
    camelcase_and_underscore_tokenizer,
)
from grub.pycode import search_documented_attributes



"""Searching for available pypi names, with word2vec query expansion"""

import re
from functools import lru_cache, cached_property

import numpy as np

from py2store.slib.s_zipfile import FileStreamsOfZip
from py2store.base import Stream
from py2store import groupby


def line_to_raw_word_vec(line):
    word, vec = line.split(maxsplit=1)
    return word.decode(), vec


class WordVecStream(Stream):
    _obj_of_data = line_to_raw_word_vec


class StreamsOfZip(FileStreamsOfZip):
    def _obj_of_data(self, data):
        return line_to_raw_word_vec(data)


def word_and_vecs(fp):
    #     fin = io.open(fname, 'r', encoding='utf-8', newline='\n', errors='ignore')

    # consume the first line (n_lines, n_dims) not yielded
    n_lines, n_dims = map(int, fp.readline().decode().split())
    for line in fp:
        tok, *vec = line.decode().rstrip().split(' ')
        yield tok, tuple(map(float, vec))


@lru_cache
def get_html():
    simple_index_url = 'https://pypi.org/simple'
    try:
        from graze.base import graze

        age_threshold = 7 * 24 * 60 * 60  # one week
        return graze(simple_index_url, max_age=age_threshold)
    except ModuleNotFoundError:
        from urllib.request import urlopen

        with urlopen(simple_index_url) as f:
            return f.read()


_parse_names_from_html = re.compile('<a href="/simple/.+/">(.+)</a>').findall


def get_distributions(html=None):
    html = html or get_html()
    return _parse_names_from_html(html.decode())


def _get_distributions_old_version_using_xml(html=None):
    from xml.etree import ElementTree
    from io import BytesIO

    html = html or get_html()
    tree = ElementTree.parse(BytesIO(html))
    return [a.text for a in tree.iter('a')]


# from py2store import lazyprop
class Pypi:
    @cached_property
    def pypi_words(self):
        return set(get_distributions())

    def is_available(self, word):
        return word not in self.pypi_words

    def available_and_not(self, words):
        return groupby(words, key=self.is_available)

    def live_is_available(self, pkg_name):
        """Check if a package name is available, but live (directly on pypi, not a cache)"""
        import urllib

        try:
            with urllib.request.urlopen(f'https://pypi.org/project/{pkg_name}') as u:
                return False
        except urllib.error.HTTPError as e:
            return True  # if url is invalid, package exists


class Search:
    """
    Example:

    ```
    zip_filepath = '/D/Dropbox/_odata/misc/wiki-news-300d-1M-subword.vec.zip'

    import pandas as pd
    df = pd.read_excel('/Users/twhalen/Downloads/pypi package names.xlsx')
    target_words = set(df.word)

    from grub.examples.pypi import Search

    s = Search(zip_filepath, search_words=target_words)
    s.search('search for the right name')
    ```
    """

    tokenizer = re.compile('\w+').findall

    def __init__(
        self,
        wordvec_zip_filepath,
        search_words,
        exclude_words='already_published',
        wordvec_name_in_zip='wiki-news-300d-1M-subword.vec',
        n_neighbors=37,
        verbose=False,
    ):
        self.wordvec_zip_filepath = wordvec_zip_filepath
        self.wordvec_name_in_zip = wordvec_name_in_zip
        self.search_words = set(search_words)
        if exclude_words:
            if exclude_words == 'already_published':
                exclude_words = set(get_distributions())
            self.search_words = self.search_words - exclude_words
        self.n_neighbors = n_neighbors
        self.verbose = verbose

    @cached_property
    def stream(self):
        return StreamsOfZip(self.wordvec_zip_filepath)

    @cached_property
    def wordvecs(self):
        if self.verbose:
            print('Gathering all the word vecs. This could take a few minutes...')
        with self.stream[self.wordvec_name_in_zip] as fp:
            all_wordvecs = dict(word_and_vecs(fp))
        return all_wordvecs

    def filtered_wordvecs(self, tok_filt):
        with self.stream[self.wordvec_name_in_zip] as fp:
            yield from filter(lambda x: tok_filt(x[0]), word_and_vecs(fp))

    def vec_matrix(self, words):
        return np.array([self.wordvecs.get(w, None) for w in words])

    def mean_vec(self, words):
        return np.mean(self.vec_matrix(words), axis=0)

    def query_to_vec(self, query):
        tokens = self.tokenizer(query)
        return self.mean_vec(tokens)

    def query_to_vec_matrix(self, query):
        tokens = self.tokenizer(query)
        return self.vec_matrix(tokens)

    @cached_property
    def knn(self):
        from sklearn.neighbors import NearestNeighbors

        taget_wv = dict(self.filtered_wordvecs(lambda x: x in self.search_words))
        X = np.array(list(taget_wv.values()))

        knn = NearestNeighbors(n_neighbors=self.n_neighbors, metric='cosine').fit(X)
        knn.words = np.array(list(taget_wv.keys()))
        return knn

    def search(self, query):
        query_vec = self.query_to_vec(query)
        r_dist, r_idx = self.knn.kneighbors(query_vec.reshape(1, -1))
        return self.knn.words[r_idx]
